# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
from earthkit.data.translators import Translator


class StrTranslator(Translator):
    def __init__(self, data, *args, **kwargs):
        self.data = data.to_string(*args, **kwargs)


def translator(data, cls, *args, **kwargs):
    if cls in [str, "string"]:
        return StrTranslator(data, *args, **kwargs)

    return None
